#include <stdio.h>
#include <time.h>
#include <stdlib.h>

/* CPU_MIN: minimum amount of CPU the benchmark has to clock up */
const  double CPU_MIN = 2.0;

void usage()
{
  printf(" Usage:\n");
  printf("bench <matrix.mtx>\n");
}
main(int argc, char *argv[])
{
  FILE* fp;
  int n,m,nz;
  int *ia, *ja, *ic, *jc;
  int i,ii,jj;
  double *a, *c;
  int ind_base, nz1;
  int *irn,*jcn, *nrow;
  double *val;
  clock_t clock_sta,clock_sto;
  double avg;
  enum {MAXLINE = 100000};
  char line[MAXLINE+1];
  /* number of runs to clock up the CPU to exceed CPU_MIN */
  int nrun;
  long flops;
  double tcomp;

  if (argc > 1) {
    if ( (fp = fopen(argv[1],"r")) == NULL) {
      printf(" can not open file %s\n",argv[1]);
      usage();
      return 0;
    }

    /* carry on doing IO until enough CPU is clocked */
    for (nrun = 1;;nrun++){
      /* get rid of comments */
      do {
	fgets(line, MAXLINE, fp);
      } while (line[0] == '%');


      /* read in he row and columb sizes and the number of nonzeros */
      sscanf(line,"%d %d %d",&n, &m, &nz) ;


      /* read in the matrix in .mtx format*/
      irn = malloc((nz)*sizeof(int));
      if (irn == NULL) {
	printf("fail allocating for irn\n");
	return 1;
      }
      jcn = malloc((nz)*sizeof(int));
      if (jcn == NULL) {
	printf("fail allocating for jcn\n");
	return 1;
      }
      val = malloc((nz)*sizeof(double));
      if (val == NULL) {
	printf("fail allocating for val\n");
	return 1;
      }

      if (nrun == 1) clock_sta = clock();
      for (i = 0; i < nz; i++) {
	fscanf(fp,"%d %d",&irn[i],&jcn[i]);
      }
      clock_sto = clock();
      if ( ((clock_sto - clock_sta)/(double) CLOCKS_PER_SEC) > CPU_MIN) break;
      rewind(fp);
      free(irn); free(jcn); free(val);
    } /* until enough CPU is clocked */
    fclose(fp);



    for (i = 0; i < nz; i++) {
      val[i] = 1.0;
    }
    printf("=========== C benchmark for sparse multiplication =============\n");
    printf(" case name = %s\n",argv[1]);
    printf(" problem size = %d\n",n);
    printf(" nz = %d\n",nz);
    printf(" time_IO = %g averaged over %d runs\n",
	   (clock_sto - clock_sta)/(double) CLOCKS_PER_SEC/(double) nrun,nrun);

    ia = malloc((n+1)*sizeof(int));
    if (ia == NULL) {
      printf("fail allocating for ia\n");
      return 1;
    }
    ja = malloc((nz)*sizeof(int));
    if (ja == NULL) {
      printf("fail allocating for ja\n");
      return 1;
    }
    a = malloc((nz)*sizeof(double));
    if (a == NULL) {
      printf("fail allocating for a\n");
      return 1;
    }

    /* convert to compact sparse row format */    
    nrow = malloc((n+1)*sizeof(int));
    for (i = 0; i < n+1; i++) {
      nrow[i] = 0;
    }
    for (i = 0; i < nz; i++) {
      nrow[irn[i]-1] = nrow[irn[i]-1] + 1;/* -1: offset for C arrays */
    }

    ia[0] = 1;/* [0]: offset for C arrays */
    for (i = 0; i < n; i++) {
      ia[i+1] = ia[i] + nrow[i];
    }

    for (i = 0; i < n+1; i++) {
      nrow[i] = ia[i] - 1; /* -1: offset for C arrays */
    }
    
    for (i = 0; i < nz; i++) {
      ii = irn[i]-1;/* -1: offset for C arrays */
      jj = jcn[i];
      ja[nrow[ii]] = jj;
      a[nrow[ii]] = val[i];
      nrow[ii] = nrow[ii] + 1;
    }
    free(irn);
    free(jcn);
    free(val);
    free(nrow);

    /* ---- COMP start ----- */
    clock_sta = clock();
    for (nrun = 1;;nrun++){
      /* work out the size of the multiplication */
      ind_base = 1;
      
      nz1 = spmatmul_size(&n,&m,ia,ja,ia,ja,&ind_base);
      
      ic = malloc((n+1)*sizeof(int));
      jc = malloc((nz1)*sizeof(int));
      c = malloc((nz1)*sizeof(double));
      
      /* work out the the multiplication */
      spmatmul_double(&n,&m,a,ia,ja,a,ia,ja,c,ic,jc,&ind_base);
      clock_sto = clock();
      if ( ((clock_sto - clock_sta)/(double) CLOCKS_PER_SEC) > CPU_MIN) break;
      free(c);free(ic);free(jc);
    } /* until enough CPU is clocked */


    printf(" nz1 = %d \n",nz1);
    tcomp = (clock_sto - clock_sta)/(double) CLOCKS_PER_SEC/(double) nrun;
    printf(" time_COMP = %g averaged over %d runs\n",
	   tcomp,nrun);
    flops = spmatmul_flops(&n,&m,ia,ja,ia,ja,&ind_base);
    printf(" Mflops/s = %g\n",flops/tcomp/1000000.0);
    printf(" flops/run = %ld\n",flops);
    
    /* average entry value */
    avg = 0;
    for (i = 0; i < nz1; i++) {
      avg = avg + c[i];
    }
    printf(" average entry value = %26.16g\n",avg/nz1);

    free(c);
    free(jc);
    free(ic);
  }
  else {
    usage();
  }
  return 0;
}



