#include <stdio.h>
#include <stdlib.h>

int spmatmul_size(const int* n, const int* m,
		   const int *ia,const int *ja,
		   const int *ib,const int *jb,
		   const int* ind_base)
{
  int* mask;
  int i,j,k,icol,nz,icol_add;
  const int *iaa;
  const int *jaa;


  iaa = ia;
  jaa = ja;
  
  mask = malloc((*m) * sizeof(int));
  for (i = 0; i < *m; i++) mask[i] = -1;

  /* offset the index for Fortran like indexing */
  ib -= *ind_base;
  jb -= *ind_base;
  mask -= *ind_base;

  nz = -1+*ind_base;
    
  for (i = 0; i < *n; i++) 
    {
      for (j = ia[i]; j < ia[i+1]; j++)
	{
	  for (k = ib[*jaa]; k < ib[*jaa+1]; k++)
	    {
	      icol_add = jb[k];
	      if (mask[icol_add] != i) 
		{
		  (nz)++;
		  mask[icol_add] = i; /* add mask */
		}
	    }
	  jaa++;
	}
    }

  /* set mask back to the correct address before freeing */
  mask += *ind_base;

  free(mask); 


  return nz;
}

void spmatmul_double(
		     const int* n, const int* m,
		     const double *a, const int *ia,const int *ja, 
		     const double *b, const int *ib,const int *jb, 
		     double *c, int *ic,int *jc,
		     const int* ind_base)
{
  int nz;
  int i,j,k,icol,icol_add;
  const double *aij;
  const int *neighbour;
  int *mask, *pmask;

  mask = malloc( (*m) * sizeof(int));
  pmask = mask;

  for (i = 0; i != *m; i++) *pmask++ = 0;
  /* shift the index base for fortran like indexing */
  ib -= *ind_base;
  jb -= *ind_base;
  b -= *ind_base;
  jc -= *ind_base;
  c -= *ind_base;
  mask -= *ind_base;

  aij = a;
  neighbour = ja;
  nz = -1+*ind_base;
  *ic = 1;
  for (i = 0; i != *n; i++) {
    for (j = ia[i]; j != ia[i+1]; j++){
      for (k = ib[*neighbour];k != ib[*neighbour+1];k++){ 
	icol_add = jb[k];
	icol = mask[icol_add];
	if (icol ==  0) 
	  {
	    jc[++nz] = icol_add;
	    c[nz] =  (*aij)*b[k];
	    mask[icol_add] = nz;
	  }
	else
	  {
	    c[icol] +=  (*aij)*b[k];
	  }
      }
      aij++;
      neighbour++;
    }
    for (j = *ic; j != nz + 1; j++) mask[jc[j]] = 0;
    *(++ic) = nz+1;
  }
  mask += *ind_base;
  free(mask);
  
}


long spmatmul_flops(const int* n, const int* m,
		   const int *ia,const int *ja,
		   const int *ib,const int *jb,
		   const int* ind_base)
{
  int* mask;
  int i,j,k,icol,nz,icol_add;
  const int *iaa;
  const int *jaa;
  long flops;

  iaa = ia;
  jaa = ja;
  
  mask = malloc((*m) * sizeof(int));
  for (i = 0; i < *m; i++) mask[i] = -1;

  /* offset the index for Fortran like indexing */
  ib -= *ind_base;
  jb -= *ind_base;
  mask -= *ind_base;

  nz = -1+*ind_base;
  flops = 0;

  for (i = 0; i < *n; i++) 
    {
      for (j = ia[i]; j < ia[i+1]; j++)
	{
	  for (k = ib[*jaa]; k < ib[*jaa+1]; k++)
	    {
	      icol_add = jb[k];
	      flops += 2;
	      if (mask[icol_add] != i) 
		{
		  (nz)++;
		  mask[icol_add] = i; /* add mask */
		}
	    }
	  jaa++;
	}
    }

  /* set mask back to the correct address before freeing */
  mask += *ind_base;

  free(mask); 


  return flops+nz;
}


